package org.dspace.app.webui;

import org.apache.log4j.Logger;
import org.dspace.app.webui.util.Authenticate;
import org.dspace.app.webui.util.JSPManager;
import org.dspace.core.ConfigurationManager;
import org.dspace.core.Context;
import org.dspace.core.LogManager;
import org.dspace.eperson.EPerson;

import javax.net.ssl.*;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.security.cert.X509Certificate;
import java.util.Properties;

/**
 * Authenticator for Central Authentication Service (CAS).
 *
* @author Nordija A/S (www.nordija.com) for Center of Knowledge Technology (www.cvt.dk)
 */
public class CASAuthenticator extends SimpleAuthenticator
{
  /**
   * log4j category
   */
  private static Logger log = Logger.getLogger(CASAuthenticator.class);

  public void startAuthentication(Context context, HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException
  {
    if (request.getParameter("bypasscas") != null)
    {
      //Just use simple authentication
      super.startAuthentication(context, request, response);
    }
    else
    {
      final String ticket = request.getParameter("ticket");
      final String service = request.getRequestURL().toString();

      log.info(LogManager.getHeader(context, "login", " ticket: " + ticket));
      log.info(LogManager.getHeader(context, "login", "service: " + service));

      if (ticket != null)
      {
        try
        {
          // Determine CAS validation URL
          String validate = ConfigurationManager.getProperty("cas.validate.url");
          log.info(LogManager.getHeader(context, "login", "CAS validate:  " + validate));
          if (validate == null)
          {
            throw new ServletException("No CAS validation URL specified. You need to set property 'cas.validate.url'");
          }

          // Validate ticket (it is assumed that CAS validator returns an email)
          final String email = validate(service, ticket, validate);
          if (email == null)
          {
            throw new ServletException("Ticket '" + ticket + "' is not valid");
          }

          // Locate the eperson
          EPerson eperson = EPerson.findByEmail(context, email);

          // Verify the password
          if (eperson != null && eperson.canLogIn())
          {
            // e-mail address corresponds to active account
            if (eperson.getRequireCertificate())
            {
              // they must use a certificate
              JSPManager.showJSP(request, response, "/error/require-certificate.jsp");
              return;
            }

            // Logged in OK.
            Authenticate.loggedIn(context, request, eperson);

            log.info(LogManager.getHeader(context, "login", "type=cas"));

            // resume previous request
            Authenticate.resumeInterruptedRequest(request, response);

            return;
          }

          // If we reach here, supplied email/password was duff.
          log.info(LogManager.getHeader(context,
              "failed_login",
              "email=" + email));
          JSPManager.showJSP(request, response, "/login/incorrect.jsp");

        } catch (Exception e)
        {
          log.error("Unexpected exception caught", e);
          throw new ServletException(e);
        }
      }
      else
      {
        // Determine CAS server URL
        final String authServer = ConfigurationManager.getProperty("cas.server.url");

        log.info("CAS server:  " + authServer);
        if (authServer == null)
        {
          throw new ServletException("No CAS server specified. You need to set property 'cas.server.url'");
        }

        // Redirect to CAS server
        response.sendRedirect(response.encodeRedirectURL(authServer + "?service=" + service));
      }
    }
  }

  /**
   * Returns the NetID of the owner of the given ticket, or null if the
   * ticket isn't valid.
   *
   * @param service the service ID for the application validating the
   *                ticket
   * @param ticket  the opaque service ticket (ST) to validate
   */
  public static String validate(String service, String ticket, String validateURL)
      throws IOException
  {
    Properties p = System.getProperties();
    p.put("java.protocol.handler.pkgs",
        "com.sun.net.ssl.internal.www.protocol");
    System.setProperties(p);                // < 1.2 compatible
    URL u = new URL(validateURL + "?ticket=" + ticket + "&service=" + service);
    BufferedReader in = new BufferedReader(new InputStreamReader(u.openStream()));
    if (in == null)
    {
      return null;
    }
    else
    {
      String line1 = in.readLine();
      log.info("CAS validation - first line:  " + line1);
      String line2 = in.readLine();
      log.info("CAS validation - second line: " + line2);
      if (line1.equals("no"))
        return null;
      else
        return line2;
    }
  }
}
