package org.dspace.webservice;

import org.apache.log4j.Logger;
import org.dspace.authorize.AuthorizeException;
import org.dspace.core.Context;
import org.dspace.core.LogManager;
import org.dspace.core.Utils;
import org.dspace.eperson.EPerson;
import org.dspace.eperson.Group;

import javax.xml.namespace.QName;
import javax.xml.rpc.soap.SOAPFaultException;
import java.sql.SQLException;

/**
 * Web service implementation
 *
 * @author Nordija A/S (www.nordija.com) for Center of Knowledge Technology (www.cvt.dk)
 */
public class DSpaceImpl implements DSpace {
  /**
   * log4j category
   */
  private static Logger log = Logger.getLogger(DSpaceImpl.class);

  public void updateEPerson(String username, String password, String email, EPersonField[] fields, EPersonGroup[] includedInGroups, EPersonGroup[] excludedFromGroups)
  {
    logUpdateEPerson(email, fields, includedInGroups, excludedFromGroups);

    Context c = null;

    try
    {
      c = obtainContext(username, password);

      internalUpdateEPerson(c, email, fields, includedInGroups, excludedFromGroups);

      c.complete();
    }
    catch (SQLException e)
    {
      final String faultCode = "F01";
      final String faultString = "Internal server error";
      logAndThrowException(faultCode, faultString, e);
    }
    catch (AuthorizeException e)
    {
      final String faultCode = "F02";
      final String faultString = "Could not authorize " + username;
      logAndThrowException(faultCode, faultString, e);
    }
    finally
    {
      // Roll back and free database connection if an exception was thrown
      if (c != null && c.isValid())
      {
        c.abort();
      }
    }
  }

  public void updateEPersons(String username, String password, EPersonUpdate[] epersonUpdate) {
    log.info("updateEPersons(..) called (number of updates is "+epersonUpdate.length+")");

    Context c = null;

    try
    {
      c = obtainContext(username, password);

      for (int i = 0; i < epersonUpdate.length; i++) {
        log.info("Updating "+(i+1)+" of "+epersonUpdate.length+" users");
        EPersonUpdate ePersonUpdate = epersonUpdate[i];
        internalUpdateEPerson(c, ePersonUpdate.getEmail(), ePersonUpdate.getFields(), ePersonUpdate.getIncludedInGroups(), ePersonUpdate.getExcludedFromGroups());
      }

      c.complete();
    }
    catch (SQLException e)
    {
      final String faultCode = "F01";
      final String faultString = "Internal server error";
      logAndThrowException(faultCode, faultString, e);
    }
    catch (AuthorizeException e)
    {
      final String faultCode = "F02";
      final String faultString = "Could not authorize " + username;
      logAndThrowException(faultCode, faultString, e);
    }
    finally
    {
      // Roll back and free database connection if an exception was thrown
      if (c != null && c.isValid())
      {
        c.abort();
      }
    }
  }

  private void internalUpdateEPerson(Context c, String email, EPersonField[] fields, EPersonGroup[] includedInGroups, EPersonGroup[] excludedFromGroups)
  {
    try
    {
      // Locate the eperson, create if user does not exist
      EPerson eperson = EPerson.findByEmail(c, email);
      if (eperson == null)
      {
        eperson = EPerson.create(c);
        eperson.setEmail(email);
        eperson.setPassword(email);
        eperson.setCanLogIn(true);
        eperson.setSelfRegistered(false);
        eperson.setRequireCertificate(false);
        if (log.isDebugEnabled()) {
          log.debug(LogManager.getHeader(c, "updateEPerson", "User="+email+" has been created"));
        }
      }

      // Update fields on eperson
      for (int i = 0; i < fields.length; i++)
      {
        final EPersonField field = fields[i];
        try
        {
          eperson.setMetadata(field.getName(), field.getValue());
          if (log.isDebugEnabled()) {
            log.debug(LogManager.getHeader(c, "updateEPerson", "User="+email+",field="+field.getName()+",value="+field.getValue()));
          }
        } catch (IllegalArgumentException e)
        {
          throw new FieldNotExistException(field);
        }
      }
      eperson.update(); // Persist changes to db (not committed yet)

      // Update groups to be included in
      for (int i = 0; i < includedInGroups.length; i++)
      {
        final EPersonGroup group = includedInGroups[i];
        Group dspaceGrp = Group.findByName(c, group.getName());

        //If group does not exist throw exception
        if (dspaceGrp == null)
        {
          throw new GroupNotExistException(group);
        }

        dspaceGrp.addMember(eperson);
        dspaceGrp.update(); // Persist changes to db (not committed yet)
        if (log.isDebugEnabled()) {
          log.debug(LogManager.getHeader(c, "updateEPerson", "User "+email+" has been included in group "+group.getName()));
        }
      }

      // Update groups to be excluded from
      for (int i = 0; i < excludedFromGroups.length; i++)
      {
        final EPersonGroup group = excludedFromGroups[i];
        final Group dspaceGrp = Group.findByName(c, group.getName());
        if (dspaceGrp == null)
        {
          log.warn(LogManager.getHeader(c, "updateEPerson", "Cannot exclude "+email+" from group "+group.getName()+" because group does not exist"));
        }
        else
        {
          dspaceGrp.removeMember(eperson);
          dspaceGrp.update(); // Persist changes to db (not committed yet)
          if (log.isDebugEnabled()) {
            log.debug(LogManager.getHeader(c, "updateEPerson", "User "+email+" has been excluded from group "+group.getName()));
          }
        }
      }

      log.info("Successfully update eperson with email " + eperson.getEmail());

    } catch (SOAPFaultException e)
    {
      throw e;
    }
    catch (FieldNotExistException e)
    {
      final String faultCode = "F03";
      final String faultString = "Field '"+e.getField().getName()+"' does not exist";
      logAndThrowException(faultCode, faultString, e);
    }
    catch (GroupNotExistException e)
    {
      final String faultCode = "F04";
      final String faultString = "Group '"+e.getGroup().getName()+"' does not exist";
      logAndThrowException(faultCode, faultString, e);
    }
    catch (Exception e)
    {
      final String faultCode = "F00";
      final String faultString = "Unexpected error";
      logAndThrowException(faultCode, faultString, e);
    }
  }

  private void logAndThrowException(final String faultCode, final String faultString, Throwable e)
  {
    log.error(faultCode+", "+faultString+"\n"+Utils.getStackTrace(e));
    throw new SOAPFaultException(new QName(faultCode), faultString, null, null);
  }

  private Context obtainContext(String username, String password) throws SQLException, AuthorizeException
  {
    // No context for this request yet
    final Context c = new Context();

    EPerson eperson = EPerson.findByEmail(c, username);

    if (eperson == null)
    {
      throw new AuthorizeException("User '"+username+"' does not exist");
    }

    if (!eperson.checkPassword(password))
    {
      throw new AuthorizeException("Incorrect password for user '" + username + "'");
    }
    if (!eperson.canLogIn())
    {
      throw new AuthorizeException("User '" + username + "' is not authorized to log in");
    }
    c.setCurrentUser(eperson);

    //Check that user is an administrator
    if (!Group.isMember(c, 1))
    {
      throw new AuthorizeException("User '" + username + "' does not have administrator rights");
    }

    return c;
  }

  /**
   * Logging method
   */
  private static void logUpdateEPerson(String email, EPersonField[] fields, EPersonGroup[] includedInGroups, EPersonGroup[] excludedFromGroups) {
    log.info("updateEPerson("+email+",..) called");

    if (log.isDebugEnabled()) {

      final StringBuffer msg = new StringBuffer("email: ");
      msg.append(email).append("\nfields:\n");
      for (int i = 0; i < fields.length; i++) {
        EPersonField field = fields[i];
        msg.append("\t").append(field.getName()).append("=").append(field.getValue()).append("\tclass=").append(field.getClass().getName()).append("\n");
      }
      msg.append(email).append("\nincluded in groups:\n");
      for (int i = 0; i < includedInGroups.length; i++) {
        EPersonGroup group = includedInGroups[i];
        msg.append("\t").append(group.getName()).append("\tclass=").append(group.getClass().getName()).append("\n");
      }
      msg.append(email).append("\nexcluded from groups:\n");
      for (int i = 0; i < excludedFromGroups.length; i++) {
        EPersonGroup group = excludedFromGroups[i];
        msg.append("\t").append(group.getName()).append("\tclass=").append(group.getClass().getName()).append("\n");
      }

      msg.append("file.encoding: ").append(System.getProperty("file.encoding")).append("\n");
      msg.append("user.language: ").append(System.getProperty("user.language")).append("\n");
      msg.append("  user.region: ").append(System.getProperty("user.region")).append("\n");

      log.debug(msg);
    }
  }
}
